@@ -118,9 +118,9 @@ import org.hsqldb.persist.Logger;
  */
 public class Database {
 
-    int            databaseID;
-    private String sType;
-    private String sName;
+    int    databaseID;
+    String sType;
+    String sName;
 
 // loosecannon1@users 1.7.2 patch properties on the JDBC URL
     private HsqlProperties urlProperties;
@@ -152,9 +152,9 @@ public class Database {
 
     // schema invarient objects
     private HashMap        hAlias;
-    private UserManager    userManager;
-    private GranteeManager granteeManager;
-    HsqlNameManager        nameManager;
+    public UserManager     userManager;
+    public GranteeManager  granteeManager;
+    public HsqlNameManager nameManager;
 
     // session related objects
     public SessionManager     sessionManager;
@@ -179,7 +179,8 @@ public class Database {
      *  Constructs a new Database object.
      *
      * @param type is the type of the database: "mem", "file", "res"
-     * @param path is the canonical path to the database files
+     * @param path is the fiven path to the database files
+     * @param name is the combination of type and canonical path
      * @param props property overrides placed on the connect URL
      * @exception  HsqlException if the specified name and path
      *      combination is illegal or unavailable, or the database files the
@@ -236,7 +237,7 @@ public class Database {
                 System.out.println("Exception");
             }
         } else {
-            fileaccess = new FileUtil();
+            fileaccess = FileUtil.getDefaultInstance();
         }
 
         shutdownOnNoConnection = urlProperties.getProperty("shutdown",
@@ -269,41 +270,37 @@ public class Database {
         setState(DATABASE_OPENING);
 
         try {
-            User sysUser;
-
             databaseProperties = new HsqlDatabaseProperties(this);
-            isNew = sType == DatabaseURL.S_MEM
+            isNew = !DatabaseURL.isFileBasedDatabaseType(sType)
                     ||!databaseProperties.checkFileExists();
 
             if (isNew && urlProperties.isPropertyTrue("ifexists")) {
-                throw Trace.error(Trace.DATABASE_NOT_EXISTS, sType + sPath);
+                throw Trace.error(Trace.DATABASE_NOT_EXISTS, sName);
             }
 
             databaseProperties.load();
             databaseProperties.setURLProperties(urlProperties);
             compiledStatementManager.reset();
 
+            nameManager           = new HsqlNameManager();
             granteeManager        = new GranteeManager(this);
             userManager           = new UserManager(this);
             hAlias                = Library.getAliasMap();
-            nameManager           = new HsqlNameManager();
             schemaManager         = new SchemaManager(this);
             bReferentialIntegrity = true;
-            sysUser               = userManager.getSysUser();
-            sessionManager        = new SessionManager(this, sysUser);
+            sessionManager        = new SessionManager(this);
             txManager             = new TransactionManager(this);
             collation             = new Collation();
             dbInfo = DatabaseInformation.newDatabaseInformation(this);
 
             databaseProperties.setDatabaseVariables();
 
-            if (sType != DatabaseURL.S_MEM) {
+            if (DatabaseURL.isFileBasedDatabaseType(sType)) {
                 logger.openLog(this);
             }
 
             if (isNew) {
-                sessionManager.getSysSession(
-                    null, false).sqlExecuteDirectNoPreChecks(
+                sessionManager.getSysSession().sqlExecuteDirectNoPreChecks(
                     "CREATE USER SA PASSWORD \"\" ADMIN");
                 logger.synchLogForce();
             }
@@ -629,44 +626,6 @@ public class Database {
         }
     }
 
-// boucherb@users - patch 1.7.2 - system change number support
-// fredt@users - system change numbers utilised
-
-    /** last statement level change number - not externally settable */
-    private long dbSCN = 0;
-
-    /** last statement level change number for DDL statements - unused */
-    private long ddlSCN = 0;
-
-    /** last statement level change number for DML statements - used for all statements */
-    private long dmlSCN = 0;
-
-    synchronized long getSCN() {
-        return dbSCN;
-    }
-
-    private synchronized void setSCN(long l) {
-        dbSCN = l;
-    }
-
-    private synchronized long nextSCN() {
-
-        dbSCN++;
-
-        return dbSCN;
-    }
-
-    synchronized long getDMLSCN() {
-        return dmlSCN;
-    }
-
-    synchronized long nextDMLSCN() {
-
-        dmlSCN = nextSCN();
-
-        return dmlSCN;
-    }
-
     private synchronized void setState(int state) {
         dbState = state;
     }
