@@ -71,9 +71,11 @@ import java.util.Locale;
 import org.hsqldb.HsqlNameManager.HsqlName;
 import org.hsqldb.lib.ArrayUtil;
 import org.hsqldb.lib.HashMap;
+import org.hsqldb.lib.HashMappedList;
 import org.hsqldb.lib.HsqlArrayList;
 import org.hsqldb.lib.IntValueHashMap;
 import org.hsqldb.store.ValuePool;
+import org.hsqldb.lib.HashSet;
 
 // fredt@users 20020130 - patch 497872 by Nitin Chauhan - reordering for speed
 // fredt@users 20020215 - patch 1.7.0 by fredt - support GROUP BY with more than one column
@@ -123,8 +125,10 @@ class Parser {
     private Database  database;
     private Tokenizer tokenizer;
     private Session   session;
+    private String    sSchema;
     private String    sTable;
     private String    sToken;
+    private boolean   wasQuoted;
     private Object    oData;
     private int       iType;
     private int       iToken;
@@ -270,7 +274,8 @@ class Parser {
                                  || predicateType == Expression.QUERY
                                  || predicateType == Expression.IN
                                  || predicateType == Expression.ALL
-                                 || predicateType == Expression.ANY;
+                                 || predicateType == Expression.ANY
+                                 || predicateType == Expression.SELECT;
         Select s = parseSelect(brackets, canHaveOrder, false, limitWithOrder,
                                true);
 
@@ -607,9 +612,8 @@ class Parser {
 
         if (brackets > 0 && token.equals(Token.T_CLOSEBRACKET)) {
             closebrackets = true;
-            brackets -= Parser.parseCloseBrackets(tokenizer, brackets - 1)
-                        + 1;
-            token = tokenizer.getString();
+            brackets      -= parseCloseBrackets(brackets - 1) + 1;
+            token         = tokenizer.getString();
         }
 
         select.unionDepth = brackets;
@@ -645,7 +649,7 @@ class Parser {
 
             if (tokenizer.isGetThis(Token.T_OPENBRACKET)) {
                 openbracket = true;
-                brackets    += Parser.parseOpenBrackets(tokenizer) + 1;
+                brackets    += parseOpenBrackets() + 1;
             }
 
             tokenizer.getThis(Token.T_SELECT);
@@ -856,71 +860,77 @@ class Parser {
                                           HsqlArrayList vfilter)
                                           throws HsqlException {
 
-        int           len     = vfilter.size();
-        TableFilter[] filters = new TableFilter[len];
+        int           colcount;
+        TableFilter[] filters = new TableFilter[vfilter.size()];
 
         vfilter.toArray(filters);
 
         select.tFilter = filters;
 
         // expand [table.]* columns
-        len = vcolumn.size();
+        colcount = vcolumn.size();
 
-        for (int i = 0; i < len; i++) {
-            Expression e = (Expression) (vcolumn.get(i));
+        for (int pos = 0; pos < colcount; ) {
+            Expression e = (Expression) (vcolumn.get(pos));
 
-            if (e.getType() == Expression.ASTERIX) {
-                int    current = i;
-                Table  table   = null;
-                String n       = e.getTableName();
+            if (e.getType() == Expression.ASTERISK) {
+                vcolumn.remove(pos);
 
-                for (int t = 0; t < filters.length; t++) {
-                    TableFilter f = filters[t];
+                colcount = vcolumn.size();
 
-//                    e.resolve(f);
-                    e.resolveTables(f);
+                String tablename = e.getTableName();
 
-                    if (n != null &&!n.equals(f.getName())) {
-                        continue;
+                if (tablename == null) {
+                    for (int i = 0; i < filters.length; i++) {
+                        pos      = addFilterColumns(filters[i], vcolumn, pos);
+                        colcount = vcolumn.size();
+                    }
+                } else {
+                    TableFilter f = e.findTableFilter(filters);
+
+                    if (f == null) {
+                        throw Trace.error(Trace.TABLE_NOT_FOUND, tablename);
                     }
 
-                    table = f.getTable();
-
-                    int col = table.getColumnCount();
-
-                    for (int c = 0; c < col; c++) {
-                        Expression ins = new Expression(f.getName(),
-                                                        table.getColumn(c));
-
-                        vcolumn.add(current++, ins);
-
-                        // now there is one element more to parse
-                        len++;
-                    }
+                    pos      = addFilterColumns(f, vcolumn, pos);
+                    colcount = vcolumn.size();
                 }
-
-                Trace.check(table != null, Trace.TABLE_NOT_FOUND, n);
-
-                // minus the asterix element
-                len--;
-
-                vcolumn.remove(current);
-            } else /* if (e.getType() == Expression.COLUMN) */ {
+            } else {
                 if (e.getFilter() == null) {
-                    for (int f = 0; f < filters.length; f++) {
-                        e.resolveTables(filters[f]);
+                    for (int i = 0; i < filters.length; i++) {
+                        e.resolveTables(filters[i]);
                     }
                 }
+
+                pos++;
             }
         }
 
-        for (int i = 0; i < len; i++) {
+        for (int i = 0; i < colcount; i++) {
             Expression e = (Expression) (vcolumn.get(i));
 
             e.resolveTypes(session);
         }
 
-        select.iResultLen = len;
+        select.iResultLen = colcount;
+    }
+
+    /**
+     * Add all columns of a table filter to list of columns
+     */
+    static int addFilterColumns(TableFilter filter, HsqlArrayList columnList,
+                                int position) {
+
+        Table table = filter.getTable();
+        int   count = table.getColumnCount();
+
+        for (int i = 0; i < count; i++) {
+            Expression e = new Expression(filter, table.getColumn(i));
+
+            columnList.add(position++, e);
+        }
+
+        return position;
     }
 
     /**
@@ -1021,7 +1031,7 @@ class Parser {
             tokenizer.back();
         }
 
-        return new TableFilter(table, alias, false);
+        return new TableFilter(table, alias, null, false);
     }
 
     /**
@@ -1035,13 +1045,13 @@ class Parser {
     private TableFilter parseTableFilter(boolean outerjoin)
     throws HsqlException {
 
-        Table       t = null;
-        SubQuery    sq;
-        TableFilter tf;
-        String      sAlias = null;
+        Table          t          = null;
+        SubQuery       sq         = null;
+        String         sAlias     = null;
+        HashMappedList columnList = null;
 
         if (tokenizer.isGetThis(Token.T_OPENBRACKET)) {
-            int brackets = Parser.parseOpenBrackets(tokenizer);
+            int brackets = parseOpenBrackets();
 
             tokenizer.getThis(Token.T_SELECT);
 
@@ -1083,15 +1093,29 @@ class Parser {
         } else if (token.equals(Token.T_AS)
                    &&!tokenizer.wasQuotedIdentifier()) {
             sAlias = tokenizer.getSimpleName();
+
+            if (tokenizer.isGetThis(Token.T_OPENBRACKET)) {
+                tokenizer.back();
+
+                columnList = parseColumnList();
+            }
         } else if (tokenizer.wasSimpleName()) {
             sAlias = token;
+
+            if (tokenizer.isGetThis(Token.T_OPENBRACKET)) {
+                tokenizer.back();
+
+                columnList = parseColumnList();
+            }
         } else {
             tokenizer.back();
         }
 
-        tf = new TableFilter(t, sAlias, outerjoin);
+        if (columnList != null && t.getColumnCount() != columnList.size()) {
+            throw Trace.error(Trace.COLUMN_COUNT_DOES_NOT_MATCH);
+        }
 
-        return tf;
+        return new TableFilter(t, sAlias, columnList, outerjoin);
     }
 
     /**
@@ -1261,7 +1285,7 @@ class Parser {
                 int brackets = 0;
 
                 if (iToken == Expression.OPEN) {
-                    brackets += Parser.parseOpenBrackets(tokenizer) + 1;
+                    brackets += parseOpenBrackets() + 1;
 
                     read();
                 }
@@ -1433,7 +1457,7 @@ class Parser {
         int        brackets = 0;
 
         if (iToken == Expression.OPEN) {
-            brackets += Parser.parseOpenBrackets(tokenizer) + 1;
+            brackets += parseOpenBrackets() + 1;
 
             read();
         }
@@ -1491,7 +1515,7 @@ class Parser {
         int        brackets = 0;
 
         if (iToken == Expression.OPEN) {
-            brackets += Parser.parseOpenBrackets(tokenizer) + 1;
+            brackets += parseOpenBrackets() + 1;
 
             read();
         }
@@ -1683,7 +1707,7 @@ class Parser {
                 break;
             }
             case Expression.SELECT : {
-
+/*
                 // accept ORDRY BY with LIMIT
                 Select select = parseSelect(0, false, false, true, true);
 
@@ -1692,7 +1716,11 @@ class Parser {
                 SubQuery sq = new SubQuery();
 
                 sq.select = select;
-                r         = new Expression(sq);
+ */
+                SubQuery sq = parseSubquery(0, null, false,
+                                            Expression.SELECT);
+
+                r = new Expression(sq);
 
                 read();
 
@@ -1706,7 +1734,7 @@ class Parser {
                 break;
             }
             case Expression.MULTIPLY : {
-                r = new Expression(sTable, (String) null);
+                r = new Expression(sSchema, sTable, (String) null);
 
                 read();
 
@@ -1907,7 +1935,7 @@ class Parser {
             r.setDataType(typeNr);
         }
 
-        r = new Expression(Expression.CONVERT, r, typeNr, length, scale);
+        r = new Expression(r, typeNr, length, scale);
 
         read();
         readThis(Expression.CLOSE);
@@ -1921,7 +1949,7 @@ class Parser {
     private Expression readColumnExpression() throws HsqlException {
 
         String     name = sToken;
-        Expression r    = new Expression(sTable, name);
+        Expression r    = new Expression(sTable, name, wasQuoted);
 
         read();
 
@@ -1929,7 +1957,7 @@ class Parser {
             String   javaName = database.getJavaName(name);
             Function f        = new Function(name, javaName, false);
 
-            session.check(javaName, UserManager.ALL);
+            session.check(javaName);
 
             int len = f.getArgCount();
             int i   = 0;
@@ -1950,7 +1978,6 @@ class Parser {
 
             readThis(Expression.CLOSE);
 
-            // TODO: Maybe allow AS <alias> here
             r = new Expression(f);
         } else {
             String javaName = (String) simpleFunctions.get(name);
@@ -2026,6 +2053,14 @@ class Parser {
 
         while (true) {
             Expression current = readOr();
+
+            if (leaf != null && iToken == Expression.CLOSE) {
+                readThis(Expression.CLOSE);
+                leaf.setLeftExpression(current);
+
+                break;
+            }
+
             Expression condition = new Expression(Expression.IS_NULL,
                                                   current, null);
             Expression alternatives = new Expression(Expression.ALTERNATIVE,
@@ -2041,12 +2076,6 @@ class Parser {
 
             leaf = alternatives;
 
-            if (iToken == Expression.CLOSE) {
-                readThis(Expression.CLOSE);
-
-                break;
-            }
-
             readThis(Expression.COMMA);
         }
 
@@ -2283,7 +2312,8 @@ class Parser {
      */
     private void read() throws HsqlException {
 
-        sToken = tokenizer.getString();
+        sToken    = tokenizer.getString();
+        wasQuoted = tokenizer.wasQuotedIdentifier();
 
         if (tokenizer.wasValue()) {
             iToken = Expression.VALUE;
@@ -2293,13 +2323,16 @@ class Parser {
             iToken = Expression.COLUMN;
             sTable = null;
         } else if (tokenizer.wasLongName()) {
-            sTable = tokenizer.getLongNameFirst();
+            sSchema = tokenizer.getLongNamePre();
+            sTable  = tokenizer.getLongNameFirst();
 
             if (sToken.equals(Token.T_MULTIPLY)) {
                 iToken = Expression.MULTIPLY;
             } else {
                 iToken = Expression.COLUMN;
             }
+        } else if (tokenizer.wasParameter()) {
+            iToken = Expression.PARAM;
         } else if (sToken.length() == 0) {
             iToken = Expression.END;
         } else {
@@ -2526,9 +2559,8 @@ class Parser {
 
         Expression expression = parseExpression();
         CompiledStatement cs = new CompiledStatement(session, database,
-            session.currentSchema, expression, getParameters());
-
-        cs.subqueries = getSortedSubqueries();
+            session.currentSchema, expression, getSortedSubqueries(),
+            getParameters());
 
         return cs;
     }
@@ -2555,9 +2587,8 @@ class Parser {
         }
 
         CompiledStatement cs = new CompiledStatement(session, database,
-            session.currentSchema, tableFilter, condition, getParameters());
-
-        cs.subqueries = getSortedSubqueries();
+            session.currentSchema, tableFilter, condition,
+            getSortedSubqueries(), getParameters());
 
         return cs;
     }
@@ -2565,42 +2596,34 @@ class Parser {
     private void getInsertColumnValueExpressions(Table t, Expression[] acve,
             int len) throws HsqlException {
 
-        boolean    enclosed;
-        String     token;
-        Expression cve;
-        int        i;
-
-        enclosed = false;
-        i        = 0;
-
         tokenizer.getThis(Token.T_OPENBRACKET);
 
-        for (; i < len; i++) {
-            cve = parseExpression();
+        for (int i = 0; i < len; i++) {
+            Expression columnValExpression = parseExpression();
 
-//            cve.resolve(null);
-            cve.resolveTables(null);
-            cve.resolveTypes(session);
+            columnValExpression.resolveTables(null);
+            columnValExpression.resolveTypes(session);
 
-            acve[i] = cve;
-            token   = tokenizer.getSimpleToken();
+            acve[i] = columnValExpression;
+
+            String token = tokenizer.getSimpleToken();
 
             if (token.equals(Token.T_COMMA)) {
                 continue;
             }
 
             if (token.equals(Token.T_CLOSEBRACKET)) {
-                enclosed = true;
-
-                break;
+                if (i == len - 1) {
+                    return;
+                } else {
+                    break;
+                }
             }
 
             tokenizer.throwUnexpected();
         }
 
-        if (!enclosed || i != len - 1) {
-            throw Trace.error(Trace.COLUMN_COUNT_DOES_NOT_MATCH);
-        }
+        throw Trace.error(Trace.COLUMN_COUNT_DOES_NOT_MATCH);
     }
 
     /**
@@ -2626,7 +2649,7 @@ class Parser {
         columnMap       = table.getColumnMap();
         len             = table.getColumnCount();
 
-        int brackets = Parser.parseOpenBrackets(tokenizer);
+        int brackets = parseOpenBrackets();
 
         token = tokenizer.getString();
 
@@ -2669,14 +2692,13 @@ class Parser {
                 CompiledStatement cs =
                     new CompiledStatement(session.currentSchema, table,
                                           columnMap, acve, columnCheckList,
+                                          getSortedSubqueries(),
                                           getParameters());
 
-                cs.subqueries = getSortedSubqueries();
-
                 return cs;
             }
             case Token.OPENBRACKET : {
-                brackets = Parser.parseOpenBrackets(tokenizer) + 1;
+                brackets = parseOpenBrackets() + 1;
 
                 tokenizer.getThis(Token.T_SELECT);
             }
@@ -2692,9 +2714,8 @@ class Parser {
 
                 CompiledStatement cs = new CompiledStatement(session,
                     database, session.currentSchema, table, columnMap,
-                    columnCheckList, select, getParameters());
-
-                cs.subqueries = getSortedSubqueries();
+                    columnCheckList, select, getSortedSubqueries(),
+                    getParameters());
 
                 return cs;
             }
@@ -2725,9 +2746,8 @@ class Parser {
         }
 
         CompiledStatement cs = new CompiledStatement(session, database,
-            session.currentSchema, select, getParameters());
-
-        cs.subqueries = getSortedSubqueries();
+            session.currentSchema, select, getSortedSubqueries(),
+            getParameters());
 
         return cs;
     }
@@ -2797,33 +2817,84 @@ class Parser {
 
         CompiledStatement cs = new CompiledStatement(session, database,
             session.currentSchema, tableFilter, colList, exprList, condition,
-            getParameters());
-
-        cs.subqueries = getSortedSubqueries();
+            getSortedSubqueries(), getParameters());
 
         return cs;
     }
 
-    static int parseOpenBrackets(Tokenizer t) throws HsqlException {
+    int parseOpenBracketsSelect() throws HsqlException {
+
+        int count = parseOpenBrackets();
+
+        tokenizer.getThis(Token.T_SELECT);
+
+        return count;
+    }
+
+    int parseOpenBrackets() throws HsqlException {
 
         int count = 0;
 
-        while (t.isGetThis(Token.T_OPENBRACKET)) {
+        while (tokenizer.isGetThis(Token.T_OPENBRACKET)) {
             count++;
         }
 
         return count;
     }
 
-    static int parseCloseBrackets(Tokenizer t,
-                                  int limit) throws HsqlException {
+    int parseCloseBrackets(int limit) throws HsqlException {
 
         int count = 0;
 
-        while (count < limit && t.isGetThis(Token.T_CLOSEBRACKET)) {
+        while (count < limit && tokenizer.isGetThis(Token.T_CLOSEBRACKET)) {
             count++;
         }
 
         return count;
     }
+
+    HashMappedList parseColumnList() throws HsqlException {
+        return processColumnList(tokenizer, false);
+    }
+
+    static HashMappedList processColumnList(Tokenizer tokenizer,
+            boolean acceptAscDesc) throws HsqlException {
+
+        HashMappedList list;
+        String         token;
+
+        list = new HashMappedList();
+
+        tokenizer.getThis(Token.T_OPENBRACKET);
+
+        while (true) {
+            token = tokenizer.getSimpleName();
+
+            boolean result = list.add(token, null);
+
+            if (!result) {
+                throw Trace.error(Trace.COLUMN_ALREADY_EXISTS, token);
+            }
+
+            token = tokenizer.getSimpleToken();
+
+            if (acceptAscDesc
+                    && (token.equals(Token.T_DESC)
+                        || token.equals(Token.T_ASC))) {
+                token = tokenizer.getSimpleToken();    // OJ: eat it up
+            }
+
+            if (token.equals(Token.T_COMMA)) {
+                continue;
+            }
+
+            if (token.equals(Token.T_CLOSEBRACKET)) {
+                break;
+            }
+
+            throw Trace.error(Trace.UNEXPECTED_TOKEN, token);
+        }
+
+        return list;
+    }
 }
