@@ -175,7 +175,7 @@ class Select {
 
         // replace the aliases with expressions
         for (int i = iResultLen; i < exprColumns.length; i++) {
-            if (exprColumns[i].exprType == Expression.COLUMN) {
+            if (exprColumns[i].getType() == Expression.COLUMN) {
                 if (exprColumns[i].joinedTableColumnIndex == -1) {
                     boolean descending = exprColumns[i].isDescending();
 
@@ -203,6 +203,30 @@ class Select {
         }
     }
 
+    /**
+     * Converts the types of the columns in set operations to those in the first
+     * Select.
+     */
+    void resolveUnionColumnTypes() throws HsqlException {
+
+        if (unionSelect != null) {
+            if (unionSelect.iResultLen != iResultLen) {
+                throw Trace.error(Trace.COLUMN_COUNT_DOES_NOT_MATCH);
+            }
+
+            for (int i = 0; i < iResultLen; i++) {
+                Expression e = exprColumns[i];
+
+                if (!e.isTypeEqual(unionSelect.exprColumns[i])) {
+                    unionSelect.exprColumns[i] =
+                        new Expression(unionSelect.exprColumns[i],
+                                       e.getDataType(), e.getColumnSize(),
+                                       e.getColumnScale());
+                }
+            }
+        }
+    }
+
     /**
      * Sets the types of all the expressions that have so far resolved.
      *
@@ -430,9 +454,14 @@ class Select {
             }
         }
 
+        // selected columns
         checkAggregateOrGroupByColumns(0, iResultLen);
+
+        // having columns
         checkAggregateOrGroupByColumns(groupByEnd, orderByStart);
-        checkAggregateOrGroupByColumns(orderByStart, orderByEnd);
+
+        // order by columns
+        checkAggregateOrGroupByOrderColumns(orderByStart, orderByEnd);
         prepareSort();
 
         simpleLimit = (isDistinctSelect == false && isGrouped == false
@@ -492,7 +521,8 @@ class Select {
 
             if (sortUnion) {
                 sortResult(session, r);
-                r.trimResult(getLimitStart(), getLimitCount(maxrows));
+                r.trimResult(getLimitStart(session),
+                             getLimitCount(session, maxrows));
             }
         }
 
@@ -581,10 +611,11 @@ class Select {
         }
     }
 
-    int getLimitStart() throws HsqlException {
+    int getLimitStart(Session session) throws HsqlException {
 
         if (limitCondition != null) {
-            Integer limit = (Integer) limitCondition.getArg().getValue(null);
+            Integer limit =
+                (Integer) limitCondition.getArg().getValue(session);
 
             if (limit != null) {
                 return limit.intValue();
@@ -599,12 +630,13 @@ class Select {
      * finds cases where the result does not have to be fully built and
      * returns an adjusted rowCount with LIMIT params.
      */
-    int getLimitCount(int rowCount) throws HsqlException {
+    int getLimitCount(Session session, int rowCount) throws HsqlException {
 
         int limitCount = 0;
 
         if (limitCondition != null) {
-            Integer limit = (Integer) limitCondition.getArg2().getValue(null);
+            Integer limit =
+                (Integer) limitCondition.getArg2().getValue(session);
 
             if (limit != null) {
                 limitCount = limit.intValue();
@@ -622,10 +654,10 @@ class Select {
      * translate the rowCount into total number of rows needed from query,
      * including any rows skipped at the beginning
      */
-    int getMaxRowCount(int rowCount) throws HsqlException {
+    int getMaxRowCount(Session session, int rowCount) throws HsqlException {
 
-        int limitStart = getLimitStart();
-        int limitCount = getLimitCount(rowCount);
+        int limitStart = getLimitStart(session);
+        int limitCount = getLimitCount(session, rowCount);
 
         if (!simpleLimit) {
             rowCount = Integer.MAX_VALUE;
@@ -651,7 +683,7 @@ class Select {
             prepareResult(session);
         }
 
-        Result r = buildResult(session, getMaxRowCount(rowCount));
+        Result r = buildResult(session, getMaxRowCount(session, rowCount));
 
         // the result is perhaps wider (due to group and order by)
         // so use the visible columns to remove duplicates
@@ -661,7 +693,8 @@ class Select {
 
         if (!sortUnion) {
             sortResult(session, r);
-            r.trimResult(getLimitStart(), getLimitCount(rowCount));
+            r.trimResult(getLimitStart(session),
+                         getLimitCount(session, rowCount));
         }
 
         return r;
@@ -724,13 +757,32 @@ class Select {
                     continue;
                 }
 
-                if (isDistinctSelect) {
-                    throw Trace.error(
-                        Trace.INVALID_ORDER_BY_IN_DISTINCT_SELECT, exp);
-                } else {
-                    throw Trace.error(Trace.NOT_IN_AGGREGATE_OR_GROUP_BY,
-                                      exp);
+                throw Trace.error(Trace.NOT_IN_AGGREGATE_OR_GROUP_BY, exp);
+            }
+        }
+    }
+
+    private void checkAggregateOrGroupByOrderColumns(int start,
+            int end) throws HsqlException {
+
+        checkAggregateOrGroupByColumns(start, end);
+
+        if (start < end && isDistinctSelect) {
+            HsqlArrayList colExps = new HsqlArrayList();
+
+            for (int i = start; i < end; i++) {
+                exprColumns[i].collectInGroupByExpressions(colExps);
+            }
+
+            for (int i = 0, size = colExps.size(); i < size; i++) {
+                Expression exp = (Expression) colExps.get(i);
+
+                if (isSimilarIn(exp, 0, iResultLen)) {
+                    continue;
                 }
+
+                throw Trace.error(Trace.INVALID_ORDER_BY_IN_DISTINCT_SELECT,
+                                  exp);
             }
         }
     }
@@ -754,8 +806,6 @@ class Select {
                    || allColumnsAreDefinedIn(exp, groupColumnNames);
         } else if (isAggregated) {
             return exp.canBeInAggregate();
-        } else if (isDistinctSelect) {
-            return isSimilarIn(exp, 0, iResultLen);
         } else {
             return true;
         }
@@ -1057,29 +1107,41 @@ class Select {
 
     boolean isResolved = false;
 
+    /**
+     * @todo - post 1.8.0 - review resolve and check resolve -
+     * determine if isResolved is specific to main query or the full set including UNION
+     *
+     */
     boolean resolveAll(Session session, boolean check) throws HsqlException {
 
-        boolean result = true;
-
         if (isResolved) {
             return true;
         }
 
         resolve(session);
 
-        result = result && checkResolved(check);
+        isResolved = checkResolved(check);
 
         if (unionSelect != null) {
             if (unionSelect.iResultLen != iResultLen) {
                 throw Trace.error(Trace.COLUMN_COUNT_DOES_NOT_MATCH);
             }
 
-            unionSelect.resolveAll(session, check);
+            for (int i = 0; i < iResultLen; i++) {
+                Expression e = exprColumns[i];
+
+                if (!e.isTypeEqual(unionSelect.exprColumns[i])) {
+                    unionSelect.exprColumns[i] =
+                        new Expression(unionSelect.exprColumns[i],
+                                       e.getDataType(), e.getColumnSize(),
+                                       e.getColumnScale());
+                }
+            }
+
+            isResolved &= unionSelect.resolveAll(session, check);
         }
 
-        isResolved = result;
-
-        return result;
+        return isResolved;
     }
 
     boolean isResolved() {
