@@ -39,6 +39,7 @@ import org.hsqldb.HsqlException;
 import org.hsqldb.Trace;
 import org.hsqldb.lib.FileAccess;
 import org.hsqldb.lib.FileUtil;
+import org.hsqldb.lib.SimpleLog;
 import org.hsqldb.lib.StopWatch;
 import org.hsqldb.lib.Storage;
 import org.hsqldb.lib.ZipUnzipFile;
@@ -62,21 +63,22 @@ import org.hsqldb.store.BitMap;
  */
 public class DataFileCache {
 
-    private FileAccess fa;
+    protected FileAccess fa;
 
     // flags
     public static final int FLAG_ISSAVED = 2;
     public static final int FLAG_ROWINFO = 3;
 
     // file format fields
-    static final int LONG_EMPTY_SIZE   = 4;     // empty space size
-    static final int LONG_FREE_POS_POS = 12;    // where iFreePos is saved
-    static final int FLAGS_POS         = 28;
-    static final int INITIAL_FREE_POS  = 32;
+    static final int LONG_EMPTY_SIZE      = 4;     // empty space size
+    static final int LONG_FREE_POS_POS    = 12;    // where iFreePos is saved
+    static final int LONG_EMPTY_INDEX_POS = 20;    // empty space index
+    static final int FLAGS_POS            = 28;
+    static final int INITIAL_FREE_POS     = 32;
 
     //
     DataFileBlockManager     freeBlocks;
-    private static final int FREE_BLOCKS_COUNT = 512;
+    private static final int initIOBufferSize = 256;
 
     //
     protected String   fileName;
@@ -107,19 +109,15 @@ public class DataFileCache {
     //
     protected Storage dataFile;
     protected long    fileFreePosition;
-    protected int     maxCacheSize;             // number of Rows
-    protected long    maxCacheBytes;            // number of bytes
+    protected int     maxCacheSize;                // number of Rows
+    protected long    maxCacheBytes;               // number of bytes
+    protected int     maxFreeBlocks;
     protected Cache   cache;
 
-    public DataFileCache(Database db, String fileName,
-                         String backupName) throws HsqlException {
+    public DataFileCache(Database db,
+                         String baseFileName) throws HsqlException {
 
-        this.fileName       = fileName;
-        this.backupFileName = backupName;
-        database            = db;
-        fa                  = database.getFileAccess();
-
-        initParams();
+        initParams(db, baseFileName);
 
         cache = new Cache(this);
     }
@@ -127,18 +125,29 @@ public class DataFileCache {
     /**
      * initial external parameters are set here.
      */
-    protected void initParams() throws HsqlException {
+    protected void initParams(Database database,
+                              String baseFileName) throws HsqlException {
 
-        int cacheScale = database.getProperties().getIntegerProperty(
+        HsqlDatabaseProperties props = database.getProperties();
+
+        fileName       = baseFileName + ".data";
+        backupFileName = baseFileName + ".backup";
+        this.database  = database;
+        fa             = database.getFileAccess();
+
+        int cacheScale = props.getIntegerProperty(
             HsqlDatabaseProperties.hsqldb_cache_scale, 14, 8, 18);
-        int cacheSizeScale = database.getProperties().getIntegerProperty(
+        int cacheSizeScale = props.getIntegerProperty(
             HsqlDatabaseProperties.hsqldb_cache_size_scale, 10, 6, 20);
+        int cacheFreeCountScale = props.getIntegerProperty(
+            HsqlDatabaseProperties.hsqldb_cache_free_count_scale, 9, 6, 12);
 
         cacheFileScale = database.getProperties().getIntegerProperty(
             HsqlDatabaseProperties.hsqldb_cache_file_scale, 1);
 
-        Trace.printSystemOut("cache_scale: " + cacheScale);
-        Trace.printSystemOut("cache_size_scale: " + cacheSizeScale);
+        if (cacheFileScale != 1) {
+            cacheFileScale = 8;
+        }
 
         cacheReadonly = database.isFilesReadOnly();
 
@@ -148,7 +157,8 @@ public class DataFileCache {
         maxCacheSize    = lookupTableLength * 3;
         maxCacheBytes   = maxCacheSize * avgRowBytes;
         maxDataFileSize = cacheFileScale == 1 ? Integer.MAX_VALUE
-                                              : (long) Integer.MAX_VALUE * 8;
+                                              : (long) Integer.MAX_VALUE * 4;
+        maxFreeBlocks   = 1 << cacheFreeCountScale;
         dataFile        = null;
     }
 
@@ -160,6 +170,8 @@ public class DataFileCache {
 
         fileFreePosition = 0;
 
+        database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "start");
+
         try {
             boolean preexists = database.isFilesInJar();
             long    freesize  = 0;
@@ -208,8 +220,8 @@ public class DataFileCache {
             String skey =
                 database.getURLProperties().getProperty("storage_key");
 
-            dataFile = ScaledRAFile.newScaledRAFile(fileName, readonly,
-                    fileType, cname, skey);
+            dataFile = ScaledRAFile.newScaledRAFile(database, fileName,
+                    readonly, fileType, cname, skey);
 
             if (preexists) {
                 dataFile.seek(FLAGS_POS);
@@ -231,15 +243,24 @@ public class DataFileCache {
                 }
             } else {
                 fileFreePosition = INITIAL_FREE_POS;
+
+                dataFile.seek(LONG_FREE_POS_POS);
+                dataFile.writeLong(INITIAL_FREE_POS);
+
+                // set unsaved flag;
+                dataFile.seek(FLAGS_POS);
+                dataFile.writeInt(0);
             }
 
             initBuffers();
 
             fileModified = false;
-            freeBlocks = new DataFileBlockManager(FREE_BLOCKS_COUNT,
+            freeBlocks = new DataFileBlockManager(maxFreeBlocks,
                                                   cacheFileScale, freesize);
+
+            database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "end");
         } catch (Throwable e) {
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, "failed");
             close(false);
 
             throw Trace.error(Trace.FILE_IO_ERROR, Trace.DataFileCache_open,
@@ -255,11 +276,13 @@ public class DataFileCache {
      *
      *  When false, just closes the file.
      *
-     *  When true, writes out all cached rows that have been modified and the free
-     *  position pointer for the *.data file and then closes the file.
+     *  When true, writes out all cached rows that have been modified and the
+     *  free position pointer for the *.data file and then closes the file.
      */
     public void close(boolean write) throws HsqlException {
 
+        SimpleLog appLog = database.logger.appLog;
+
         try {
             if (cacheReadonly) {
                 if (dataFile != null) {
@@ -271,36 +294,54 @@ public class DataFileCache {
 
             StopWatch sw = new StopWatch();
 
+            appLog.sendLine(SimpleLog.LOG_NORMAL,
+                            "DataFileCache.close(" + write + ") : start");
+
             if (write) {
                 cache.saveAll();
                 Trace.printSystemOut("saveAll: " + sw.elapsedTime());
+                appLog.sendLine(SimpleLog.LOG_NORMAL,
+                                "DataFileCache.close() : save data");
 
-                // set empty
-                dataFile.seek(LONG_EMPTY_SIZE);
-                dataFile.writeLong(freeBlocks.getLostBlocksSize());
+                if (fileModified || freeBlocks.isModified()) {
 
-                // set end
-                dataFile.seek(LONG_FREE_POS_POS);
-                dataFile.writeLong(fileFreePosition);
+                    // set empty
+                    dataFile.seek(LONG_EMPTY_SIZE);
+                    dataFile.writeLong(freeBlocks.getLostBlocksSize());
 
-                // set saved flag;
-                dataFile.seek(FLAGS_POS);
+                    // set end
+                    dataFile.seek(LONG_FREE_POS_POS);
+                    dataFile.writeLong(fileFreePosition);
 
-                int flag = BitMap.set(0, FLAG_ISSAVED);
+                    // set saved flag;
+                    dataFile.seek(FLAGS_POS);
 
-                if (hasRowInfo) {
-                    flag = BitMap.set(flag, FLAG_ROWINFO);
+                    int flag = BitMap.set(0, FLAG_ISSAVED);
+
+                    if (hasRowInfo) {
+                        flag = BitMap.set(flag, FLAG_ROWINFO);
+                    }
+
+                    dataFile.writeInt(flag);
+                    appLog.sendLine(SimpleLog.LOG_NORMAL,
+                                    "DataFileCache.close() : flags");
+
+                    //
+                    if (dataFile.length() != fileFreePosition) {
+                        dataFile.seek(fileFreePosition);
+                    }
+
+                    appLog.sendLine(SimpleLog.LOG_NORMAL,
+                                    "DataFileCache.close() : seek end");
+                    Trace.printSystemOut("pos and flags: "
+                                         + sw.elapsedTime());
                 }
-
-                dataFile.writeInt(flag);
-
-                //
-                dataFile.seek(fileFreePosition);
-                Trace.printSystemOut("pos and flags: " + sw.elapsedTime());
             }
 
             if (dataFile != null) {
                 dataFile.close();
+                appLog.sendLine(SimpleLog.LOG_NORMAL,
+                                "DataFileCache.close() : close");
 
                 dataFile = null;
 
@@ -314,7 +355,7 @@ public class DataFileCache {
                 fa.removeElement(backupFileName);
             }
         } catch (Throwable e) {
-            database.logger.appLog.logContext(e);
+            appLog.logContext(e, null);
 
             throw Trace.error(Trace.FILE_IO_ERROR, Trace.DataFileCache_close,
                               new Object[] {
@@ -323,42 +364,17 @@ public class DataFileCache {
         }
     }
 
-    void postClose(boolean keep) throws HsqlException {
-
-        if (cacheReadonly) {
-            return;
-        }
-
-        try {
-            if (keep) {
-                database.getProperties().setProperty(
-                    HsqlDatabaseProperties.hsqldb_cache_version,
-                    HsqlDatabaseProperties.VERSION_STRING_1_7_0);
-                database.getProperties().save();
-
-                if (fileModified) {
-                    backup();
-                }
-            } else {
-                fa.removeElement(backupFileName);
-                deleteOrResetFreePos(database, fileName);
-            }
-        } catch (IOException e) {
-            throw new HsqlException(
-                e, Trace.getMessage(Trace.GENERAL_IO_ERROR),
-                Trace.GENERAL_IO_ERROR);
-        }
-    }
-
     protected void initBuffers() {
 
         if (rowOut == null
-                || ((RowOutputBinary) rowOut).getBuffer().length > 256) {
+                || ((RowOutputBinary) rowOut).getBuffer().length
+                   > initIOBufferSize) {
             rowOut = new RowOutputBinary(256);
         }
 
         if (rowIn == null
-                || ((RowInputBinary) rowIn).getBuffer().length > 256) {
+                || ((RowInputBinary) rowIn).getBuffer().length
+                   > initIOBufferSize) {
             rowIn = new RowInputBinary(new byte[256]);
         }
     }
@@ -376,55 +392,40 @@ public class DataFileCache {
             return;
         }
 
+        database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "start");
+
         try {
-            boolean        wasNio = dataFile.wasNio();
+            boolean wasNio = dataFile.wasNio();
+
+            cache.saveAll();
+
             DataFileDefrag dfd = new DataFileDefrag(database, this, fileName);
 
             dfd.process();
             close(false);
-            Trace.printSystemOut("closed old cache");
-
-            // first attemp to delete
-            fa.removeElement(fileName);
-
-            if (wasNio) {
-                System.gc();
-
-                if (fa.isStreamElement(fileName)) {
-                    fa.removeElement(fileName);
-
-                    if (fa.isStreamElement(fileName)) {
-                        fa.renameElement(fileName, fileName + ".old");
-
-                        File oldfile = new File(fileName + ".old");
-
-                        FileUtil.deleteOnExit(oldfile);
-                    }
-                }
-            }
-
-// oj@openofice.org - change to file access api
-            fa.renameElement(fileName + ".new", fileName);
-            backup();
+            deleteFile(wasNio);
+            renameDataFile();
+            backupFile();
             database.getProperties().setProperty(
                 HsqlDatabaseProperties.hsqldb_cache_version,
                 HsqlDatabaseProperties.VERSION_STRING_1_7_0);
             database.getProperties().save();
-            initParams();
             cache.clear();
 
             cache = new Cache(this);
 
             open(cacheReadonly);
             dfd.updateTableIndexRoots();
-            Trace.printSystemOut("opened cache");
+            dfd.updateTransactionRowIDs();
         } catch (Throwable e) {
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, null);
 
             throw new HsqlException(
                 e, Trace.getMessage(Trace.GENERAL_IO_ERROR),
                 Trace.GENERAL_IO_ERROR);
         }
+
+        database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "end");
     }
 
     /**
@@ -432,7 +433,9 @@ public class DataFileCache {
      * Removes the row from the cache data structures.
      * Adds the file space for the row to the list of free positions.
      */
-    public void remove(int i, PersistentStore store) throws HsqlException {
+    public synchronized void remove(int i,
+                                    PersistentStore store)
+                                    throws IOException {
 
         CachedObject r    = release(i);
         int          size = r == null ? getStorageSize(i)
@@ -441,6 +444,9 @@ public class DataFileCache {
         freeBlocks.add(i, size);
     }
 
+    public synchronized void removePersistence(int i,
+            PersistentStore store) throws IOException {}
+
     /**
      * Allocates file space for the row. <p>
      *
@@ -471,7 +477,7 @@ public class DataFileCache {
         return i;
     }
 
-    public void add(CachedObject object) throws IOException {
+    public synchronized void add(CachedObject object) throws IOException {
 
         int size = object.getRealSize(rowOut);
 
@@ -494,7 +500,7 @@ public class DataFileCache {
      * For a CacheObject that had been previously released from the cache.
      * A new version is introduced, using the preallocated space for the object.
      */
-    public void restore(CachedObject object) throws IOException {
+    public synchronized void restore(CachedObject object) throws IOException {
 
         int i = object.getPos();
 
@@ -506,24 +512,15 @@ public class DataFileCache {
         }
     }
 
-    public synchronized int getStorageSize(int i) throws HsqlException {
+    public synchronized int getStorageSize(int i) throws IOException {
 
-        try {
-            CachedObject value = cache.get(i);
+        CachedObject value = cache.get(i);
 
-            if (value != null) {
-                return value.getStorageSize();
-            }
-
-            return readSize(i);
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
-
-            throw Trace.error(Trace.DATA_FILE_ERROR,
-                              Trace.DataFileCache_makeRow, new Object[] {
-                e, fileName
-            });
+        if (value != null) {
+            return value.getStorageSize();
         }
+
+        return readSize(i);
     }
 
     public synchronized CachedObject get(int i, PersistentStore store,
@@ -558,8 +555,7 @@ public class DataFileCache {
 
             return object;
         } catch (IOException e) {
-            database.logger.appLog.logContext("" + cache + " pos: " + i);
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, fileName + " get pos: " + i);
 
             throw Trace.error(Trace.DATA_FILE_ERROR,
                               Trace.DataFileCache_makeRow, new Object[] {
@@ -568,18 +564,19 @@ public class DataFileCache {
         }
     }
 
-    RowInputInterface getRaw(int i) throws IOException {
+    synchronized RowInputInterface getRaw(int i) throws IOException {
         return readObject(i);
     }
 
-    protected int readSize(int pos) throws IOException {
+    protected synchronized int readSize(int pos) throws IOException {
 
         dataFile.seek((long) pos * cacheFileScale);
 
         return dataFile.readInt();
     }
 
-    protected RowInputInterface readObject(int pos) throws IOException {
+    protected synchronized RowInputInterface readObject(int pos)
+    throws IOException {
 
         dataFile.seek((long) pos * cacheFileScale);
 
@@ -591,33 +588,39 @@ public class DataFileCache {
         return rowIn;
     }
 
-    public CachedObject release(int i) {
+    public synchronized CachedObject release(int i) {
         return cache.release(i);
     }
 
-    /**
-     * This is called internally when old rows need to be removed from the
-     * cache.
-     */
-    protected void saveRows(CachedObject[] rows, int offset,
-                            int count) throws IOException {
+    protected synchronized void saveRows(CachedObject[] rows, int offset,
+                                         int count) throws IOException {
 
-        for (int i = offset; i < offset + count; i++) {
-            CachedObject r = rows[i];
+        try {
+            for (int i = offset; i < offset + count; i++) {
+                CachedObject r = rows[i];
 
-            saveRow(r);
+                saveRow(r);
 
-            rows[i] = null;
+                rows[i] = null;
+            }
+        } catch (IOException e) {
+            database.logger.appLog.logContext(e, null);
+
+            throw e;
+        } catch (Throwable e) {
+            database.logger.appLog.logContext(e, null);
+
+            throw new IOException(e.toString());
+        } finally {
+            initBuffers();
         }
-
-        initBuffers();
     }
 
     /**
      * Writes out the specified Row. Will write only the Nodes or both Nodes
      * and table row data depending on what is not already persisted to disk.
      */
-    public void saveRow(CachedObject row) throws IOException {
+    public synchronized void saveRow(CachedObject row) throws IOException {
 
         setFileModified();
         rowOut.reset();
@@ -632,33 +635,75 @@ public class DataFileCache {
      *
      * @throws  HsqlException
      */
-    public void backup() throws IOException {
+    void backupFile() throws IOException {
 
         try {
-            ZipUnzipFile.compressFile(fileName, backupFileName + ".new",
-                                      database.getFileAccess());
-            fa.renameElement(backupFileName + ".new", backupFileName);
+            if (fa.isStreamElement(fileName)) {
+                ZipUnzipFile.compressFile(fileName, backupFileName + ".new",
+                                          database.getFileAccess());
+            }
         } catch (IOException e) {
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, null);
 
             throw e;
         }
     }
 
+    void renameBackupFile() {
+
+        if (fa.isStreamElement(backupFileName + ".new")) {
+            fa.removeElement(backupFileName);
+            fa.renameElement(backupFileName + ".new", backupFileName);
+        }
+    }
+
+    /**
+     *  Renames the *.data.new file.
+     *
+     * @throws  HsqlException
+     */
+    void renameDataFile() {
+
+        if (fa.isStreamElement(fileName + ".new")) {
+            fa.removeElement(fileName);
+            fa.renameElement(fileName + ".new", fileName);
+        }
+    }
+
+    void deleteFile(boolean wasNio) {
+
+        // first attemp to delete
+        fa.removeElement(fileName);
+
+        if (fa.isStreamElement(fileName)) {
+            if (wasNio) {
+                System.gc();
+                fa.removeElement(fileName);
+            }
+
+            if (fa.isStreamElement(fileName)) {
+                fa.renameElement(fileName, fileName + ".old");
+
+                File oldfile = new File(fileName + ".old");
+
+                FileUtil.deleteOnExit(oldfile);
+            }
+        }
+    }
+
+    void deleteBackup() {
+        fa.removeElement(backupFileName);
+    }
+
     /**
      * This method deletes a data file or resets its free position.
      * this is used only for nio files - not OOo files
      */
-    public static void deleteOrResetFreePos(Database database,
-            String filename) {
+    static void deleteOrResetFreePos(Database database, String filename) {
 
         ScaledRAFile raFile = null;
 
-        try {
-            database.getFileAccess().removeElement(filename);
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
-        }
+        database.getFileAccess().removeElement(filename);
 
         if (database.isStoredFileAccess()) {
             return;
@@ -669,42 +714,23 @@ public class DataFileCache {
         }
 
         try {
-            raFile = new ScaledRAFile(filename, false);
+            raFile = new ScaledRAFile(database, filename, false);
 
             raFile.seek(LONG_FREE_POS_POS);
             raFile.writeLong(INITIAL_FREE_POS);
         } catch (IOException e) {
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, null);
         } finally {
             if (raFile != null) {
                 try {
                     raFile.close();
                 } catch (IOException e) {
-                    database.logger.appLog.logContext(e);
+                    database.logger.appLog.logContext(e, null);
                 }
             }
         }
     }
 
-    public static int getFlags(String filename) throws HsqlException {
-
-        try {
-            ScaledRAFile raFile =
-                (ScaledRAFile) ScaledRAFile.newScaledRAFile(filename, true,
-                    ScaledRAFile.DATA_FILE_RAF, null, null);
-
-            raFile.seek(FLAGS_POS);
-
-            int flags = raFile.readInt();
-
-            raFile.close();
-
-            return flags;
-        } catch (IOException e) {
-            throw Trace.error(Trace.DATA_FILE_ERROR);
-        }
-    }
-
     public int capacity() {
         return maxCacheSize;
     }
@@ -745,7 +771,7 @@ public class DataFileCache {
         return fileModified;
     }
 
-    protected void setFileModified() throws IOException {
+    protected synchronized void setFileModified() throws IOException {
 
         if (!fileModified) {
 
