@@ -78,6 +78,7 @@ import org.hsqldb.Trace;
 import org.hsqldb.lib.FileAccess;
 import org.hsqldb.lib.HashMap;
 import org.hsqldb.lib.Iterator;
+import org.hsqldb.lib.SimpleLog;
 import org.hsqldb.lib.ZipUnzipFile;
 import org.hsqldb.scriptio.ScriptReaderBase;
 import org.hsqldb.scriptio.ScriptWriterBase;
@@ -131,8 +132,6 @@ public class Log {
     private FileAccess             fa;
     private ScriptWriterBase       dbLogWriter;
     private String                 scriptFileName;
-    private String                 cacheFileName;
-    private String                 backupFileName;
     private String                 logFileName;
     private boolean                filesReadOnly;
     private long                   maxLogSize;
@@ -148,15 +147,7 @@ public class Log {
         properties = db.getProperties();
     }
 
-    /**
-     * When opening a database, the hsqldb.compatible_version property is
-     * used to determine if this version of the engine is equal to or greater
-     * than the earliest version of the engine capable of opening that
-     * database.<p>
-     *
-     * @throws  HsqlException
-     */
-    void open() throws HsqlException {
+    void initParams() {
 
         // Allows the user to set log size in the properties file.
         int logMegas = properties.getIntegerProperty(
@@ -170,8 +161,19 @@ public class Log {
         filesReadOnly  = database.isFilesReadOnly();
         scriptFileName = fileName + ".script";
         logFileName    = fileName + ".log";
-        cacheFileName  = fileName + ".data";
-        backupFileName = fileName + ".backup";
+    }
+
+    /**
+     * When opening a database, the hsqldb.compatible_version property is
+     * used to determine if this version of the engine is equal to or greater
+     * than the earliest version of the engine capable of opening that
+     * database.<p>
+     *
+     * @throws  HsqlException
+     */
+    void open() throws HsqlException {
+
+        initParams();
 
         int state = properties.getDBModified();
 
@@ -192,6 +194,20 @@ public class Log {
                 reopenAllTextCaches();
                 break;
 
+            case HsqlDatabaseProperties.FILES_NEW :
+                try {
+                    deleteBackup();
+                    backupData();
+                    renameNewBackup();
+                    renameNewScript();
+                    deleteLog();
+                    properties.setDBModified(
+                        HsqlDatabaseProperties.FILES_NOT_MODIFIED);
+                } catch (IOException e) {
+                    database.logger.appLog.logContext(e, null);
+                }
+
+            // continue as non-modified files
             case HsqlDatabaseProperties.FILES_NOT_MODIFIED :
 
                 /**
@@ -202,6 +218,8 @@ public class Log {
                 processScript();
 
                 if (isAnyCacheModified()) {
+                    properties.setDBModified(
+                        HsqlDatabaseProperties.FILES_MODIFIED);
                     close(false);
 
                     if (cache != null) {
@@ -211,9 +229,6 @@ public class Log {
                     reopenAllTextCaches();
                 }
                 break;
-
-            case HsqlDatabaseProperties.FILES_NEW :
-                break;
         }
 
         openLog();
@@ -234,22 +249,11 @@ public class Log {
 
         closeLog();
         deleteNewAndOldFiles();
-
-        // create '.script.new'
         writeScript(script);
         closeAllTextCaches(script);
 
         if (cache != null) {
-            cache.close(!script);
-            cache.postClose(!script);
-        }
-
-        // oj@openoffice.org - changed to file access api
-        try {
-            fa.renameElement(scriptFileName + ".new", scriptFileName);
-            fa.removeElement(logFileName);
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
+            cache.close(true);
         }
 
         properties.setProperty(HsqlDatabaseProperties.db_version,
@@ -259,6 +263,20 @@ public class Log {
             HsqlDatabaseProperties.FIRST_COMPATIBLE_VERSION);
 
         // set this one last to save the props
+        properties.setDBModified(HsqlDatabaseProperties.FILES_NEW);
+        deleteLog();
+
+        if (script) {
+            deleteBackup();
+            deleteData();
+        } else {
+            try {
+                backupData();
+                renameNewBackup();
+            } catch (IOException e) {}
+        }
+
+        renameNewScript();
         properties.setDBModified(HsqlDatabaseProperties.FILES_NOT_MODIFIED);
     }
 
@@ -283,16 +301,55 @@ public class Log {
      */
     void deleteNewAndOldFiles() {
 
-        try {
-            fa.removeElement(cacheFileName + ".old");
-            fa.removeElement(cacheFileName + ".new");
-            fa.removeElement(backupFileName + ".new");
-            fa.removeElement(scriptFileName + ".new");
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
+        fa.removeElement(fileName + ".data" + ".old");
+        fa.removeElement(fileName + ".data" + ".new");
+        fa.removeElement(fileName + ".backup" + ".new");
+        fa.removeElement(scriptFileName + ".new");
+    }
+
+    void deleteBackup() {
+        fa.removeElement(fileName + ".backup");
+    }
+
+    void deleteData() {
+        fa.removeElement(fileName + ".data");
+    }
+
+    void backupData() throws IOException {
+
+        if (fa.isStreamElement(fileName + ".data")) {
+            ZipUnzipFile.compressFile(fileName + ".data",
+                                      fileName + ".backup.new",
+                                      database.getFileAccess());
         }
     }
 
+    void renameNewBackup() {
+
+        if (fa.isStreamElement(fileName + ".backup.new")) {
+            fa.renameElement(fileName + ".backup.new", fileName + ".backup");
+        }
+    }
+
+    void renameNewScript() {
+
+        if (fa.isStreamElement(scriptFileName + ".new")) {
+            fa.renameElement(scriptFileName + ".new", scriptFileName);
+        }
+    }
+
+    void deleteNewScript() {
+        fa.removeElement(scriptFileName + ".new");
+    }
+
+    void deleteNewBackup() {
+        fa.removeElement(scriptFileName + ".backup.new");
+    }
+
+    void deleteLog() {
+        fa.removeElement(logFileName);
+    }
+
     /**
      * Checks all the caches and returns true if the modified flag is set for any
      */
@@ -315,9 +372,8 @@ public class Log {
             return;
         }
 
-        closeLog();
-        closeAllTextCaches(false);
-        reopenAllTextCaches();
+        database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "start");
+        deleteNewAndOldFiles();
 
         if (cache != null) {
             if (forceDefrag()) {
@@ -327,27 +383,29 @@ public class Log {
             if (defrag) {
                 try {
                     cache.defrag();
-                } catch (Exception e) {
-                    cache.close(true);
-                    cache.postClose(true);
-                    cache.open(false);
-                }
+                } catch (Exception e) {}
             } else {
                 cache.close(true);
-                cache.postClose(true);
+
+                try {
+                    cache.backupFile();
+                } catch (IOException e1) {
+                    deleteNewBackup();
+                    cache.open(false);
+
+                    return;
+                }
+
                 cache.open(false);
             }
         }
 
         writeScript(false);
-
-        try {
-            fa.renameElement(scriptFileName + ".new", scriptFileName);
-            fa.removeElement(logFileName);
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
-        }
-
+        properties.setDBModified(HsqlDatabaseProperties.FILES_NEW);
+        closeLog();
+        deleteLog();
+        renameNewScript();
+        renameNewBackup();
         properties.setDBModified(HsqlDatabaseProperties.FILES_MODIFIED);
 
         if (dbLogWriter == null) {
@@ -370,6 +428,8 @@ public class Log {
         } catch (IOException e) {
             throw Trace.error(Trace.FILE_IO_ERROR, logFileName);
         }
+
+        database.logger.appLog.logContext(SimpleLog.LOG_NORMAL, "end");
     }
 
     /**
@@ -380,8 +440,9 @@ public class Log {
         long megas = properties.getIntegerProperty(
             HsqlDatabaseProperties.hsqldb_defrag_limit, 200);
         long defraglimit = megas * 1024 * 1024;
+        long lostSize    = cache.freeBlocks.getLostBlocksSize();
 
-        return cache.freeBlocks.getLostBlocksSize() > defraglimit;
+        return lostSize > defraglimit;
     }
 
     /**
@@ -402,8 +463,7 @@ public class Log {
         }
 */
         if (cache == null) {
-            cache = new DataFileCache(database, cacheFileName,
-                                      backupFileName);
+            cache = new DataFileCache(database, fileName);
 
             cache.open(filesReadOnly);
         }
@@ -417,7 +477,8 @@ public class Log {
 
     void setLogSize(int megas) {
 
-        properties.setProperty("hsqldb.log_size", megas);
+        properties.setProperty(HsqlDatabaseProperties.hsqldb_log_size,
+                               String.valueOf(megas));
 
         maxLogSize = megas * 1024 * 1024;
     }
@@ -440,10 +501,11 @@ public class Log {
 
         scriptFormat = type;
 
-        properties.setProperty("hsqldb.script_format", scriptFormat);
+        properties.setProperty(HsqlDatabaseProperties.hsqldb_script_format,
+                               String.valueOf(scriptFormat));
 
         if (needsCheckpoint) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -480,7 +542,7 @@ public class Log {
         }
 
         if (maxLogSize > 0 && dbLogWriter.size() > maxLogSize) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -494,7 +556,7 @@ public class Log {
         }
 
         if (maxLogSize > 0 && dbLogWriter.size() > maxLogSize) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -508,7 +570,7 @@ public class Log {
         }
 
         if (maxLogSize > 0 && dbLogWriter.size() > maxLogSize) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -522,7 +584,7 @@ public class Log {
         }
 
         if (maxLogSize > 0 && dbLogWriter.size() > maxLogSize) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -535,7 +597,7 @@ public class Log {
         }
 
         if (maxLogSize > 0 && dbLogWriter.size() > maxLogSize) {
-            checkpoint(false);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -580,15 +642,7 @@ public class Log {
      */
     private void writeScript(boolean full) throws HsqlException {
 
-        String sNewName = scriptFileName + ".new";
-
-        try {
-            if (fa.isStreamElement(sNewName)) {
-                fa.removeElement(sNewName);
-            }
-        } catch (IOException e) {
-            database.logger.appLog.logContext(e);
-        }
+        deleteNewScript();
 
         //fredt - to do - flag for chache set index
         ScriptWriterBase scw = ScriptWriterBase.newScriptWriter(database,
@@ -616,7 +670,7 @@ public class Log {
                         true));
                 scr.close();
             }
-        } catch (IOException e) {
+        } catch (Throwable e) {
             if (scr != null) {
                 scr.close();
 
@@ -627,9 +681,17 @@ public class Log {
                 closeAllTextCaches(false);
             }
 
-            database.logger.appLog.logContext(e);
+            database.logger.appLog.logContext(e, null);
 
-            throw Trace.error(Trace.FILE_IO_ERROR, e.getMessage());
+            if (e instanceof HsqlException) {
+                throw (HsqlException) e;
+            } else if (e instanceof IOException) {
+                throw Trace.error(Trace.FILE_IO_ERROR, e.toString());
+            } else if (e instanceof OutOfMemoryError) {
+                throw Trace.error(Trace.OUT_OF_MEMORY);
+            } else {
+                throw Trace.error(Trace.GENERAL_ERROR, e.toString());
+            }
         }
     }
 
@@ -648,7 +710,7 @@ public class Log {
         long dataLength = cache.getFileFreePos();
 
         if (logLength + dataLength > cache.maxDataFileSize) {
-            checkpoint(true);
+            database.logger.needsCheckpoint = true;
         }
     }
 
@@ -669,15 +731,16 @@ public class Log {
     private void restoreBackup() throws HsqlException {
 
         // in case data file cannot be deleted, reset it
-        DataFileCache.deleteOrResetFreePos(database, cacheFileName);
+        DataFileCache.deleteOrResetFreePos(database, fileName + ".data");
 
         try {
-            ZipUnzipFile.decompressFile(backupFileName, cacheFileName,
+            ZipUnzipFile.decompressFile(fileName + ".backup",
+                                        fileName + ".data",
                                         database.getFileAccess());
         } catch (Exception e) {
             throw Trace.error(Trace.FILE_IO_ERROR, Trace.Message_Pair,
                               new Object[] {
-                backupFileName, e.getMessage()
+                fileName + ".backup", e.toString()
             });
         }
     }
